# This module should calculate verification metrics at any time of the workflow
# It should implement different verification metrics:
# - FRPSS and RPSS
# - FCRPSS and CRPSS
# - enscorr
# - bias
# - reliability diagram
# - ask Carlos which decadal metrics he is currently using

source("modules/Skill/compute_quants.R")
source("modules/Skill/compute_probs.R")
source("modules/Skill/s2s.metrics.R")
## TODO: Remove when new version of s2dv is released
source("modules/Skill/tmp/RandomWalkTest.R")
source("modules/Skill/tmp/Bias.R")
source("modules/Skill/tmp/AbsBiasSS.R")

## TODO: Implement this in the future
## Which parameter are required?
# if (!("obs" %in% ls()) || is.null(obs)) {
#   error(logger,
#         "There is no object 'obs' in the global environment or it is NULL")
# }
# if (stream == "fcst" && (!("fcst" %in% ls()) || is.null(fcst))) {
#   error(logger,
#         "There is no object 'fcst' in the global environment or it is NULL")
# }
# if (!("hcst" %in% ls()) || is.null(hcst)) {
#   error(logger,
#         "There is no object 'hcst' in the global environment or it is NULL")
# }
# if (!("metric" %in% ls()) || is.null(metric)) {
#   warn(logger,
#        "Verification metric not found and it is set as 'EnsCorr'.")
#   metric <- 'EnsCorr'
# }
# if (metric %in% c('FRPSS', 'RPSS')) {
#   metric_fun <- "veriApply"
#   metric_method <- "FairRpss"
# } else if (metric %in% c("FCRPSS", "CRPSS")) {
#   metric_fun <- "veriApply"
# } else if (metric %in% c("EnsCorr", "EnsCor")) {
#   metric_fun <- "veriApply"
#   metric_method <- "EnsCorr"
# #...
# } else {
#   error(logger, "Unknown verification metric defined in the recipe.")
#   metric_fun <- 'NotFound'
# }
# info(logger, paste("#-------------------------- ", "\n",
#                    " running Skill module ", "\n",
#                    " it can call ", metric_fun ))

compute_skill_metrics <- function(recipe, exp, obs) {
  # exp: s2dv_cube containing the hindcast
  # obs: s2dv_cube containing the observations
  # recipe: auto-s2s recipe as provided by read_yaml
  
  ## TODO: Adapt time_dims to subseasonal case
  time_dim <- 'syear'
  memb_dim <- 'ensemble'
  metrics <- tolower(recipe$Analysis$Workflow$Skill$metric)
  if (is.null(recipe$Analysis$ncores)) {
    ncores <- 1
  } else {
    ncores <- recipe$Analysis$ncores
  }
  if (is.null(recipe$Analysis$remove_NAs)) {
    na.rm = F
  } else {
    na.rm = recipe$Analysis$remove_NAs
  }
  skill_metrics <- list()
  for (metric in strsplit(metrics, ", | |,")[[1]]) {
    # Whether the fair version of the metric is to be computed
    if (metric %in% c('frps', 'frpss', 'bss10', 'bss90',
		      'fcrps', 'fcrpss')) {
      Fair <- T
    } else {
      Fair <- F
    }
    # Whether to compute correlation for the ensemble mean or for each member
    if (metric == 'corr') {
      memb <- T
    } else if (metric == 'enscorr') {
      memb <- F
    } 
    # Ranked Probability Score and Fair version
    if (metric %in% c('rps', 'frps')) {
      skill <- RPS(exp$data, obs$data, time_dim = time_dim, memb_dim = memb_dim,
          	   Fair = Fair, ncores = ncores)
      skill <- .drop_dims(skill)
      skill_metrics[[ metric ]] <- skill
    # Ranked Probability Skill Score and Fair version
    } else if (metric %in% c('rpss', 'frpss')) {
      skill <- RPSS(exp$data, obs$data, time_dim = time_dim, memb_dim = memb_dim,
          	    Fair = Fair, ncores = ncores)
      skill <- lapply(skill, function(x) {
		      .drop_dims(x)})
      skill_metrics[[ metric ]] <- skill$rpss
      skill_metrics[[ paste0(metric, "_significance") ]] <- skill$sign
    # Brier Skill Score - 10th percentile
    } else if (metric == 'bss10') {
      skill <- RPSS(exp$data, obs$data, time_dim = time_dim, 
		    memb_dim = memb_dim, prob_thresholds = 0.1, Fair = Fair,
		    ncores = ncores)
      skill <- lapply(skill, function(x) {
		      .drop_dims(x)})
      skill_metrics[[ metric ]] <- skill$rpss
      skill_metrics[[ paste0(metric, "_significance") ]] <- skill$sign
    # Brier Skill Score - 90th percentile
    } else if (metric == 'bss90') {
      skill <- RPSS(exp$data, obs$data, time_dim = time_dim,
		    memb_dim = memb_dim, prob_thresholds = 0.9, Fair = Fair,
		    ncores = ncores)
      skill <- lapply(skill, function(x) {
		      .drop_dims(x)})
      skill_metrics[[ metric ]] <- skill$rpss
      skill_metrics[[ paste0(metric, "_significance") ]] <- skill$sign
    # CRPS and FCRPS
    } else if (metric %in% c('crps', 'fcrps')) {
      skill <- CRPS(exp$data, obs$data, time_dim = time_dim,
		    memb_dim = memb_dim, Fair = Fair, ncores = ncores)
      skill <- .drop_dims(skill)
      skill_metrics[[ metric ]] <- skill
    # CRPSS and FCRPSS
    } else if (metric %in% c('crpss', 'fcrpss')) {
      skill <- CRPSS(exp$data, obs$data, time_dim = time_dim,
		     memb_dim = memb_dim, Fair = Fair, ncores = ncores)
      skill <- lapply(skill, function(x) {
		      .drop_dims(x)})
      skill_metrics[[ metric ]] <- skill$crpss
      skill_metrics[[ paste0(metric, "_significance") ]] <- skill$sign
   # Mean bias (climatology) 
   } else if (metric == 'mean_bias') {
      skill <- Bias(exp$data, obs$data, time_dim = time_dim,
		    memb_dim = memb_dim, ncores = ncores)
      skill <- .drop_dims(skill)
      skill_metrics[[ metric ]] <- skill
    # Mean bias skill score
    } else if (metric == 'mean_bias_ss') {
      skill <- AbsBiasSS(exp$data, obs$data, time_dim = time_dim,
		         memb_dim = memb_dim, ncores = ncores)
      skill <- lapply(skill, function(x) {
		      .drop_dims(x)})
      skill_metrics[[ metric ]] <- skill$biasSS
      skill_metrics[[ paste0(metric, "_significance") ]] <- skill$sign
    # Ensemble mean correlation
    } else if (metric %in% c('enscorr', 'corr')) {
      ## TODO: Implement option for Kendall and Spearman methods?
      skill <- s2dv::Corr(exp$data, obs$data, 
          		  dat_dim = 'dat', 
          		  time_dim = time_dim, 
          		  method = 'pearson', 
          		  memb_dim = memb_dim,
			  memb = memb,
          		  ncores = ncores)
      skill <- lapply(skill, function(x) {
		      .drop_dims(x)})
      skill_metrics[[ metric ]] <- skill$corr
      skill_metrics[[ paste0(metric, "_p.value") ]] <- skill$p.val
      skill_metrics[[ paste0(metric, "_conf.low") ]] <- skill$conf.lower
      skill_metrics[[ paste0(metric, "_conf.up") ]] <- skill$conf.upper
    } else if (metric == 'enssprerr') {
      # Remove ensemble dim from obs to avoid veriApply warning
      obs_noensdim <- ClimProjDiags::Subset(obs$data, "ensemble", 1,
					    drop = "selected")
      capture.output(
      skill <- easyVerification::veriApply(verifun = 'EnsSprErr',
					   fcst = exp$data,
					   obs = obs_noensdim,
					   tdim = which(names(dim(exp$data))==time_dim),
					   ensdim = which(names(dim(exp$data))==memb_dim),
					   na.rm = na.rm,
					   ncpus = ncores)
      )
      remove(obs_noensdim)
      skill <- .drop_dims(skill)
      skill_metrics[[ metric ]] <- skill
    # SpecsVerification metrics  
    } else if (grepl("specs", metric, fixed = TRUE)) {
      # Compute SpecsVerification version of the metrics
      ## Retain _specs in metric name for clarity
      metric_name <- (strsplit(metric, "_"))[[1]][1] # Get metric name
      if (!(metric_name %in% c('frpss', 'frps', 'bss10', 'bss90', 'enscorr',
			       'rpss'))) {
	## TODO: Test this scenario
	warn(recipe$Run$logger,
	     "Some of the requested metrics are not available.")
      }
      capture.output(
      skill <- Compute_verif_metrics(exp$data, obs$data,
          			     skill_metrics = metric_name,
          			     verif.dims=c("syear", "sday", "sweek"),
          			     na.rm = na.rm,
          			     ncores = ncores)
      )
      skill <- .drop_dims(skill)
      if (metric_name == "frps") {
	# Compute yearly mean for FRPS
	skill <- colMeans(skill, dims = 1)
      }
      skill_metrics[[ metric ]] <- skill
    }
  }
  info(recipe$Run$logger, "##### SKILL METRIC COMPUTATION COMPLETE #####")
  return(skill_metrics)
}

compute_probabilities <- function(recipe, data) {

  if (is.null(recipe$Analysis$ncores)) {
    ncores <- 1
  } else {
    ncores <- recipe$Analysis$ncores
  }
  
  if (is.null(recipe$Analysis$remove_NAs)) {
    na.rm = F
  } else {
    na.rm = recipe$Analysis$remove_NAs
  }
  
  named_probs <- list()
  named_quantiles <- list()
  if (is.null(recipe$Analysis$Workflow$Probabilities$percentiles)) {
    error(recipe$Run$logger, "Quantiles and probability bins have been 
	  requested, but no thresholds are provided in the recipe.")
    stop()
  } else {
    for (element in recipe$Analysis$Workflow$Probabilities$percentiles) {
      # Parse thresholds in recipe
      thresholds <- sapply(element, function (x) eval(parse(text = x)))
      quants <- compute_quants(data$data, thresholds,
			       ncores = ncores,
			       na.rm = na.rm)
      probs <- compute_probs(data$data, quants,
		             ncores = ncores,
		             na.rm = na.rm)
      for (i in seq(1:dim(quants)['bin'][[1]])) {
	named_quantiles <- append(named_quantiles,
				  list(ClimProjDiags::Subset(quants,
							     'bin', i)))
	names(named_quantiles)[length(named_quantiles)] <- paste0("percentile_",
					    as.integer(thresholds[i]*100))
      }
      for (i in seq(1:dim(probs)['bin'][[1]])) {
	if (i == 1) {
	  name_i <- paste0("prob_b", as.integer(thresholds[1]*100))
	} else if (i == dim(probs)['bin'][[1]]) {
	  name_i <- paste0("prob_a", as.integer(thresholds[i-1]*100))
	} else {
	  name_i <- paste0("prob_", as.integer(thresholds[i-1]*100), "_to_",
			   as.integer(thresholds[i]*100))
	}
	named_probs <- append(named_probs,
			      list(ClimProjDiags::Subset(probs,
							 'bin', i)))
	names(named_probs)[length(named_probs)] <- name_i
      }
    }
    named_probs <- lapply(named_probs, function(x) {.drop_dims(x)})
    named_quantiles <- lapply(named_quantiles, function(x) {.drop_dims(x)})
  }
  info(recipe$Run$logger, 
       "##### PERCENTILES AND PROBABILITY CATEGORIES COMPUTED #####")
  return(list(probs=named_probs, percentiles=named_quantiles))
}

## TODO: Replace with ClimProjDiags::Subset
.drop_dims <- function(metric_array) {
  # Drop all singleton dimensions
  metric_array <- drop(metric_array)
  # If time happened to be a singleton dimension, add it back in the array
  if (!("time" %in% names(dim(metric_array)))) {
    dim(metric_array) <- c("time" = 1, dim(metric_array))
  }
  # If array has memb_exp (Corr case), change name to 'ensemble'
  if ("exp_memb" %in% names(dim(metric_array))) {
    names(dim(metric_array))[which(names(dim(metric_array)) == 
				   "exp_memb")] <- "ensemble"
  # } else {
  #   dim(metric_array) <- c(dim(metric_array), "ensemble" = 1)
  }
  return(metric_array)
}
